import numpy as np

from woffl.flow import jetflow as jf
from woffl.flow import jetplot as jplt
from woffl.flow import outflow as of
from woffl.flow import singlephase as sp
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import Annulus, Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.resmix import ResMix


def suction_check(
    form_temp: float,
    jpump_tvd: float,
    rho_pf: float,
    ppf_surf: float,
    jpump_well: JetPump,
    tube: Pipe,
    ipr_well: InFlow,
    prop_well: ResMix,
) -> None:
    """Suction Check Choked Conditions

    The following function looks at the operating conditions of a jet pump with a choked throat
    entry. The function does no comparison of how the pump will behave based on discharge conditions
    or any other constraints. Prints results as text, creates plots of the jet pump throat entry
    and diffuser.

    Args:
        form_temp (float): Formation Temperature, deg F
        jpump_tvd (float): Jet Pump True Vertical Depth, feet
        rho_pf (float): Power Fluid Density, lbm/ft3
        ppf_surf (float): Pressure of Power Fluid at surface, psig
        jpump_well (JetPump): Jet Pump Class
        tube (Pipe): Pipe Class, used for diffuser diameter
        ipr_well (InFlow): IPR Class
        prop_well (ResMix): Reservoir conditions of the well

    Returns:
        None
    """
    psu_min, qoil_std, te_book = jf.psu_minimize(
        tsu=form_temp, ken=jpump_well.ken, ate=jpump_well.ate, ipr_su=ipr_well, prop_su=prop_well
    )
    pte, vte, rho_te, mach_te = te_book.dete_zero()

    pni = ppf_surf + sp.diff_press_static(rho_pf, jpump_tvd)
    vnz = jf.nozzle_velocity(pni, pte, jpump_well.knz, rho_pf)

    qnz_ft3s, qnz_bpd = jf.nozzle_rate(vnz, jpump_well.anz)
    wc_tm, qwat_su = jf.throat_wc(qoil_std, prop_well.wc, qnz_bpd)

    prop_tm = ResMix(wc_tm, prop_well.fgor, prop_well.oil, prop_well.wat, prop_well.gas)
    ptm = jf.throat_discharge(
        pte, form_temp, jpump_well.kth, vnz, jpump_well.anz, rho_pf, vte, jpump_well.ate, rho_te, prop_tm
    )
    vtm, pdi = jf.diffuser_discharge(ptm, form_temp, jpump_well.kdi, jpump_well.ath, tube.inn_area, qoil_std, prop_tm)

    print(f"Suction Pressure: {round(psu_min, 1)} psig")
    print(f"Oil Flow: {round(qoil_std, 1)} bopd")
    print(f"Nozzle Inlet Pressure: {round(pni, 1)} psig")
    print(f"Throat Entry Pressure: {round(pte, 1)} psig")
    print(f"Throat Discharge Pressure: {round(ptm, 1)} psig")
    print(f"Diffuser Discharge Pressure: {round(pdi, 1)} psig")
    print(f"Power Fluid Rate: {round(qnz_bpd, 1)} bwpd")
    print(f"Nozzle Velocity: {round(vnz, 1)} ft/s")
    print(f"Throat Entry Velocity: {round(vte, 1)} ft/s")

    qoil_std, te_book = jplt.throat_entry_book(psu_min, form_temp, jpump_well.ken, jpump_well.ate, ipr_well, prop_well)
    te_book.plot_te()

    vtm, di_book = jplt.diffuser_book(ptm, form_temp, jpump_well.ath, jpump_well.kdi, tube.inn_area, qoil_std, prop_tm)
    di_book.plot_di()


def discharge_check(
    surf_pres: float,
    form_temp: float,
    rho_pf: float,
    ppf_surf: float,
    jpump_well: JetPump,
    tube: Pipe,
    wellprof: WellProfile,
    ipr_well: InFlow,
    prop_well: ResMix,
) -> None:
    """Discharge Check Choked Conditions

    The following function compares what the jet pump can discharge compared to what the
    discharge pressure needs to be to lift the well. It only looks at choked conditions of
    the jet pump instead of iterating to find a zero residual. Prints results as text, creates
    plots of the jet pump throat entry and diffuser.

    Args:
        surf_pres (float): Well Head Surface Pressure, psig
        form_temp (float): Formation Temperature, deg F
        rho_pf (float): Power Fluid Density, lbm/ft3
        ppf_surf (float): Pressure of Power Fluid at surface, psig
        jpump_well (JetPump): Jet Pump Class
        tube (Pipe): Pipe Class
        wellprof (WellProfile): Well Profile Class
        ipr_well (InFlow): IPR Class
        prop_well (ResMix): Reservoir conditions of the well

    Returns:
        None
    """
    psu_min, qoil_std, te_book = jf.psu_minimize(
        tsu=form_temp, ken=jpump_well.ken, ate=jpump_well.ate, ipr_su=ipr_well, prop_su=prop_well
    )
    pte, vte, rho_te, mach_te = te_book.dete_zero()
    pni = ppf_surf + sp.diff_press_static(rho_pf, wellprof.jetpump_vd)
    vnz = jf.nozzle_velocity(pni, pte, jpump_well.knz, rho_pf)

    qnz_ft3s, qnz_bpd = jf.nozzle_rate(vnz, jpump_well.anz)
    wc_tm, qwat_su = jf.throat_wc(qoil_std, prop_well.wc, qnz_bpd)

    prop_tm = ResMix(wc_tm, prop_well.fgor, prop_well.oil, prop_well.wat, prop_well.gas)
    ptm = jf.throat_discharge(
        pte, form_temp, jpump_well.kth, vnz, jpump_well.anz, rho_pf, vte, jpump_well.ate, rho_te, prop_tm
    )
    vtm, pdi = jf.diffuser_discharge(ptm, form_temp, jpump_well.kdi, jpump_well.ath, tube.inn_area, qoil_std, prop_tm)

    md_seg, prs_ray, slh_ray = of.top_down_press(surf_pres, form_temp, qoil_std, prop_tm, tube, wellprof)

    outflow_pdi = prs_ray[-1]
    diff_pdi = pdi - outflow_pdi

    if diff_pdi >= 0:
        pdi_str = f"Well will flow choked, discharge pressure is {round(diff_pdi, 0)} psig greater than required"
    else:
        pdi_str = f"Well will NOT flow choked, discharge pressure is {round(diff_pdi, 0)} psig below the required"

    print(f"Suction Pressure: {round(psu_min, 1)} psig")
    print(f"Oil Flow: {round(qoil_std, 1)} bopd")
    print(f"Nozzle Inlet Pressure: {round(pni, 1)} psig")
    print(f"Throat Entry Pressure: {round(pte, 1)} psig")
    print(f"Throat Discharge Pressure: {round(ptm, 1)} psig")
    print(f"Required Diffuser Discharge Pressure: {round(prs_ray[-1], 1)} psig")
    print(f"Supplied Diffuser Discharge Pressure: {round(pdi, 1)} psig")
    print(pdi_str)
    print(f"Power Fluid Rate: {round(qnz_bpd, 1)} bwpd")
    print(f"Nozzle Velocity: {round(vnz, 1)} ft/s")
    print(f"Throat Entry Velocity: {round(vte, 1)} ft/s")

    # add the outflow, with the liquid holdup and pressure

    # graphing some outputs for visualization
    qsu_std, te_book = jplt.throat_entry_book(psu_min, form_temp, jpump_well.ken, jpump_well.ate, ipr_well, prop_well)
    te_book.plot_te()
    # print(te_book)
    vtm, di_book = jplt.diffuser_book(ptm, form_temp, jpump_well.ath, jpump_well.kdi, tube.inn_area, qsu_std, prop_tm)
    di_book.plot_di()
    # print(di_book)
    # te_book.plot()



"""Batch Jet Pump Runs

Contains code that is used to run multiple pumps at once to understand the
current conditions.
"""

import pandas as pd

from woffl.assembly.easypump import jetpump_wrapper

"""
This is a file to call the run_pump file from Assembly folder

Loops through a list of jet pump nozzle and throat sizes and returns a csv file with the results
Update well characterisitics in the call to the model pump function

"""


nozzles = ["9", "10", "11", "12", "13", "14"]
throats = ["X", "A", "B", "C", "D", "E"]

results = []

for nozzle in nozzles:
    for throat in throats:
        try:
            psu_solv, sonic_status, qoil_std, fwat_bwpd, qnz_bwpd, mach_te, total_wc, total_water, wellname = (
                jetpump_wrapper(
                    True,  # isSchrader?
                    pwh=201,  # WHP
                    rho_pf=62.4,  # PF density
                    ppf_surf=2419,  # PF pres
                    out_dia=4.5,  # Tubing OD
                    thick=0.5,  # tubing thickness
                    qwf=320,  # Oil rate
                    pwf=547,  # FBHP
                    res_pres=800,  # res pressure
                    form_wc=0.01,  # watercut
                    form_gor=507,  # gor
                    form_temp=72,  # form temp
                    nozzle_no=nozzle,  # nozzle size
                    throat=throat,  # nozzel area ratio with throat
                    wellname="MPB-35",
                )
            )
            result = {
                "nozzle": nozzle,
                "throat": throat,
                "psu_solv": psu_solv,
                "sonic_status": sonic_status,
                "qoil_std": qoil_std,
                "fwat_bwpd": fwat_bwpd,
                "qnz_bwpd": qnz_bwpd,
                "mach_te": mach_te,
                "total water": total_water,
                "total_wc": total_wc,
                "wellname": wellname,
            }
            results.append(result)
        except Exception as e:
            print(f"An error occurred for nozzle {nozzle} and throat {throat}: {e}")

df_results = pd.DataFrame(results)
df_sorted = df_results.sort_values(by="psu_solv", ascending=True)
# df_sorted.to_csv("modelrun_output B-35.csv")
print(df_sorted)



"""PVT Folder

Collection of classes, functions and methods for describing properties
of hydrocarbon mixtures.

ResMix (Class): Blah Blah
BlackOil (Class):
FormGas (Class):
FormWater (Class):
"""

from .blackoil import BlackOil
from .formgas import FormGas
from .formwat import FormWater
from .resmix import ResMix



"""Assembly Folder

The assembly folder is for the collection of functions and classes that are
used to analyze the entire Well Assembly. The assembly consists of an IPR, Jet Pump
Well Bore, Well Profile and Surface Constraints. Each assembly is unique and possesses
its own set of unique results to be analyzed.
"""



"""Out Flow for an Oil Well

Performs the calculations that occur inside a vertical, inclined or horizontal wellbore.
Predominate equations are two phase flow correlations. The most common being Beggs and Brill
"""

import numpy as np

from woffl.flow import singlephase as sp
from woffl.flow import twophase as tp
from woffl.geometry import forms as fm
from woffl.geometry.pipe import Annulus, Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.resmix import ResMix


def homo_diff_press(
    pin: float, tin: float, inn_dia: float, abs_ruff: float, length: float, height: float, qoil_std: float, prop: ResMix
) -> tuple[float, float, float]:
    """Homogenous Differential Pressure

    Calculate Homogenous Differential Pressure Across Wellbore / Pipe.
    Note, could have positive / negative length and height, depending on which node you
    are starting with. EG, starting at bottom and going up or starting at top and going down.

    Args:
        pin (float): Inlet Pressure, psig
        tin (float): Inlet Temperature, deg F
        inn_dia (float): Inner Diameter of Pipe, inches
        abs_ruff (float): Absolute Roughness of Pipe, inches
        length (float): Length of Pipe, feet, Positive is with flow direction
        height (float): Height Difference of Pipe, feet, Positive is upwards
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix

    Returns:
        dp (float): Differential Pressure, psid
        nslh (float): No Slip Liquid Holdup, unitless
    """
    prop = prop.condition(pin, tin)
    qtot = sum(prop.insitu_volm_flow(qoil_std))
    rho_mix = prop.rho_mix()
    visc_mix = prop.visc_mix()
    nslh = prop.nslh()

    area = sp.cross_area(inn_dia)
    vmix = sp.velocity(qtot, area)
    NRe = sp.reynolds(rho_mix, vmix, inn_dia, visc_mix)
    rel_ruff = sp.relative_roughness(inn_dia, abs_ruff)
    ff = sp.ffactor_darcy(NRe, rel_ruff)

    dp_fric = sp.diff_press_friction(ff, rho_mix, vmix, inn_dia, length)
    dp_stat = sp.diff_press_static(rho_mix, height)
    return dp_stat, dp_fric, nslh


def beggs_diff_press(
    pin: float, tin: float, inn_dia: float, abs_ruff: float, length: float, height: float, qoil_std: float, prop: ResMix
) -> tuple[float, float, float]:
    """Beggs Differential Pressure

    Calculate Beggs Differential Pressure Across Wellbore / Pipe.
    Note, could have positive / negative length and height, depending on which node you
    are starting with. EG, starting at bottom and going up or starting at top and going down.

    Args:
        pin (float): Inlet Pressure, psig
        tin (float): Inlet Temperature, deg F
        inn_dia (float): Inner Diameter of Pipe, inches
        abs_ruff (float): Absolute Roughness of Pipe, inches
        length (float): Length of Pipe, feet, Positive is with flow direction
        height (float): Height Difference of Pipe, feet, Positive is upwards
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix

    Returns:
        dp_stat (float): Static Differential Pressure, psid
        dp_fric (float): Friction Differntial Pressure, psid
        slh (float): Slip Liquid Holdup, unitless
    """
    prop = prop.condition(pin, tin)
    qoil, qwat, qgas = prop.insitu_volm_flow(qoil_std)
    rho_liq, rho_gas = prop.rho_two()
    sig_liq = prop.tension()
    nslh = prop.nslh()

    visc_mix = prop.visc_mix()
    rho_mix = prop.rho_mix()

    area = sp.cross_area(inn_dia)
    vsl, vsg, vmix = tp.velocities(qoil, qwat, qgas, area)
    NFr = tp.froude(vmix, inn_dia)
    NLv = tp.ros_nlv(vsl, rho_liq, sig_liq)  # ros liquid velocity number

    hpat, tran = tp.beggs_flow_pattern(nslh, NFr)
    incline = fm.horz_angle(length, height)
    slh = tp.beggs_holdup_inc(nslh, NFr, NLv, incline, hpat, tran)
    slh = min(slh, 1)  # liquid holdup never above one (before or after payne?)
    slh = tp.payne_correction(slh, incline)  # 1979 correction
    rho_slip = tp.density_slip(rho_liq, rho_gas, slh)
    dp_stat = tp.beggs_press_static(rho_slip, height)

    NRe = sp.reynolds(rho_mix, vmix, inn_dia, visc_mix)
    rel_ruff = sp.relative_roughness(inn_dia, abs_ruff)
    ff = sp.ffactor_darcy(NRe, rel_ruff)  # make this nomenclature consistent with beggs?
    yb = tp.beggs_yf(nslh, slh)  # NSLh, Lh
    sb = tp.beggs_sf(yb)
    fb = tp.beggs_ff(ff, sb)
    dp_fric = tp.beggs_press_friction(fb, rho_mix, vmix, inn_dia, length)
    return dp_stat, dp_fric, slh


def top_down_press(
    ptop: float, ttop: float, qoil_std: float, prop: ResMix, tubing: Pipe, wellprof: WellProfile, model: str = "beggs"
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Top Down WellBore Pressure Calculation

    Uses the specificed model to calculate the pressure gradient in a wellbore, starting
    at the top and working down to inflow / jet pump node. This is the preferred method
    of calculation, as a negative pressure cannot be created during calculations.

    Args:
        ptop (float): Pressure at top node, psig
        ttop (float): Temperature at top node, deg F
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix
        tubing (Pipe): Piping geometry inside the wellbore, Pipe
        wellprof (WellProfile): survey dimensions and location of jet pump, WellProfile
        model (str): Specify which model to use, either homo or beggs

    Returns:
        md_seg (list): Measured depth of calculated pressure
        prs_ray (list): Calculated pressure along wellbore, psig
        slh_ray (list): Liquid Holdup along wellbore, unitless
    """
    # mp_models = {'homo': homo_diff_press(), 'beggs': beggs_diff_press()}
    prs_ray = np.array([ptop])
    slh_ray = np.array([])
    md_seg, vd_seg = wellprof.outflow_spacing(100)  # space every 100'
    md_diff = np.diff(md_seg, n=1) * -1  # against flow
    vd_diff = np.diff(vd_seg, n=1) * -1  # going down piping
    n = 0
    for length, height in zip(md_diff, vd_diff):
        dp_stat, dp_fric, slh = beggs_diff_press(
            prs_ray[-1], ttop, tubing.inn_dia, tubing.abs_ruff, length, height, qoil_std, prop
        )
        pdwn = prs_ray[-1] - dp_stat - dp_fric  # dp is subtracted
        prs_ray = np.append(prs_ray, pdwn)
        slh_ray = np.append(slh_ray, slh)
        n = n + 1
    # the no slip array is going to be one shorter than the md_seg and prs_ray...
    # i'm not sure if this is problem that I should "fix" later?
    return md_seg, prs_ray, slh_ray


def bottom_up_press(
    pbot: float, tbot: float, qoil_std: float, prop: ResMix, tubing: Pipe, wellprof: WellProfile, model: str = "beggs"
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """NOT TESTED: Bottom Up WellBore Pressure Calculation

    Uses the specificed model to calculate the pressure gradient in a wellbore, starting
    at the bottom inflow / jet pump node and working up to ther surface. Problem with this
    method is that a negative pressure can be calculated, causing an error.

    Args:
        pbot (float): Pressure at bottom node, psig
        tbot (float): Temperature at bottom node, deg F
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix
        tubing (Pipe): Piping geometry inside the wellbore, Pipe
        wellprof (WellProfile): survey dimensions and location of jet pump, WellProfile
        model (str): Specify which model to use, either homo or beggs

    Returns:
        md_seg (list): Measured depth of calculated pressure
        prs_ray (list): Calculated pressure along wellbore, psig
        slh_ray (list): Liquid Holdup along wellbore, unitless
    """
    # mp_models = {'homo': homo_diff_press(), 'beggs': beggs_diff_press()}
    prs_ray = np.array([pbot])
    slh_ray = np.array([])
    md_seg, vd_seg = wellprof.outflow_spacing(100)  # space every 100'
    md_diff = np.diff(md_seg, n=1)  # with the flow
    vd_diff = np.diff(vd_seg, n=1)  # going up piping
    n = 0
    for length, height in zip(np.flip(md_diff), np.flip(vd_diff)):  # start at bottom
        dp_stat, dp_fric, slh = beggs_diff_press(
            prs_ray[-1], tbot, tubing.inn_dia, tubing.abs_ruff, length, height, qoil_std, prop
        )
        pdwn = prs_ray[-1] - dp_stat - dp_fric  # dp is subtracted
        prs_ray = np.append(prs_ray, pdwn)
        slh_ray = np.append(slh_ray, slh)
        n = n + 1
    # the no slip array is going to be one shorter than the md_seg and prs_ray...
    # i'm not sure if this is problem that I should "fix" later?
    return md_seg, prs_ray, slh_ray



import numpy as np

from woffl.flow import jetcheck as jc
from woffl.flow import jetflow as jf
from woffl.flow import jetplot as jplt
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import Annulus, Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.blackoil import BlackOil
from woffl.pvt.formgas import FormGas
from woffl.pvt.formwat import FormWater
from woffl.pvt.resmix import ResMix

# data from MPU E-41 Well Test on 11/27/2023
# only works if the command python -m tests.e41_test is used

surf_pres = 210
jpump_tvd = 4065  # feet, interpolated off well profile
rho_pf = 62.4  # lbm/ft3
ppf_surf = 3168  # psi, power fluid surf pressure 3168

# testing the jet pump code on E-41
tube = Pipe(out_dia=4.5, thick=0.5)  # E-42 tubing
case = Pipe(out_dia=6.875, thick=0.5)  # E-42 casing
ann = Annulus(inn_pipe=tube, out_pipe=case)  # define the annulus

e41_ipr = InFlow(qwf=246, pwf=1049, pres=1400)  # define an ipr

e41_jp = JetPump(nozzle_no="13", area_ratio="A", ken=0.03, kth=0.3, kdi=0.4)

mpu_oil = BlackOil.schrader_oil()  # class method
mpu_wat = FormWater.schrader_wat()  # class method
mpu_gas = FormGas.schrader_gas()  # class method

form_wc = 0.894
form_gor = 600  # formation gor
form_temp = 111
e41_res = ResMix(wc=form_wc, fgor=form_gor, oil=mpu_oil, wat=mpu_wat, gas=mpu_gas)

e42_profile = WellProfile.schrader()

jc.discharge_check(surf_pres, form_temp, rho_pf, ppf_surf, e41_jp, tube, e42_profile, e41_ipr, e41_res)
# jc.jet_check(form_temp, jpump_tvd, rho_pf, ppf_surf, e41_jp, tube, e41_ipr, e41_res)

psu_ray = np.linspace(1106, 1250, 5)
qoil_list, book_list = jplt.multi_throat_entry_books(psu_ray, form_temp, e41_jp.ken, e41_jp.ate, e41_ipr, e41_res)
jplt.multi_suction_graphs(qoil_list, book_list)



import math


class Pipe:
    def __init__(self, out_dia: float, thick: float, abs_ruff: float = 0.004) -> None:
        """Initialize a Piece of Piping

        The pipe doesn't have a length. The object will work with the profile
        class to create an appropriate length and angle of piping.

        Args:
            out_dia (float): Outside diameter of the pipe, inches
            thick (float): Thickness of the piping, inches
            abs_ruff (float): Abs Roughness, inches, Default: 0.004 inches

        Returns:
            Self
        """

        if thick > out_dia:
            raise ValueError("Pipe thickness is greater than outer diameter")

        self.out_dia = out_dia
        self.thick = thick
        self.abs_ruff = abs_ruff

        self.inn_dia = out_dia - 2 * thick  # inner diamter

    def __repr__(self):
        return f"Pipe: Out Dia {self.out_dia} inches and Inn Dia {self.inn_dia} inches"

    @property
    def inn_area(self) -> float:
        """Inner area of piping, ft2"""
        return self.area_circle(self.inn_dia) / 144

    @property
    def out_area(self) -> float:
        """Outer area of piping, ft2"""
        return self.area_circle(self.out_dia) / 144

    @staticmethod
    def area_circle(diameter):
        """Area of a Circle

        Args:
            diameter (float): diameter of a circle

        Returns:
            area (float): area of a circle
        """
        area = math.pi * (diameter**2) / 4
        return area


class Annulus:
    def __init__(self, inn_pipe: Pipe, out_pipe: Pipe) -> None:
        """Initialize a pipe annulus

        The annulus doesn't have a length. The object will work with the
        profile class to create an appropriate length and angle of annulus.

        Args:
            inn_pipe (Pipe): Inside Pipe in an annulus set up
            out_pipe (Pipe): Outside Pipe in an annulus set up

        Returns:
            Self
        """
        if inn_pipe.out_dia > out_pipe.inn_dia:
            raise ValueError("Inner pipe will not fit into the outer pipe")

        self.inn_pipe = inn_pipe
        self.out_pipe = out_pipe

    def __repr__(self):
        return f"Inner Pipe OD: {self.inn_pipe.out_dia} inches, Outer Pipe OD: {self.out_pipe.out_dia} inches"

    @property
    def ann_area(self) -> float:
        """Annulus area of annulus, ft2"""
        return self.out_pipe.inn_area - self.inn_pipe.out_area

    @property
    def hyd_dia(self) -> float:
        """Annulus hydraulic diameter, inches"""
        return self.out_pipe.inn_dia - self.inn_pipe.out_dia



"""Jet Flow Equations

Functions that are used for solving the fluid dynamics inside a jet pump. The actual
jet pump geometry is accomplished in a seperate module.
"""

import math

import numpy as np
from scipy.integrate import trapezoid

from woffl.flow import jetplot as jp
from woffl.flow import singlephase as sp
from woffl.flow.inflow import InFlow
from woffl.pvt.resmix import ResMix


def enterance_ke(ken: float, vte: float) -> float:
    """Throat Enterance Kinetic Energy

    Calculate the kinetic energy in the throat enterance.
    Add the energy lost due to friction to balance out with EE.

    Args:
        ken: Throat Entry Friction Factor, unitless
        vte: Velocity at Throat Entry, ft/s

    Returns:
        ke_te: Throat Enterance Kinetic Energy, ft2/s2
    """
    ke_te = (1 + ken) * (vte**2) / 2
    return ke_te


def incremental_ee(prs_ray: np.ndarray, rho_ray: np.ndarray) -> float:
    """Fluid Incremental Expansion Energy

    Calculate the incremental change in expansion energy for a fluid over a
    pressure change. Uses the trapezoid rule to sum the area under the density
    curve for a difference in pressure. Equal to Sum (dp/ρ). The length of the
    arrays must match and be equal or greater than length 2.

    Args:
        prs_ray (np.ndarray): Array of pressures, psig
        rho_ray (np.ndarray): Array of densitys, lbm/ft3

    Returns:
        ee_inc (float): Expansion Energy Incremental, ft2/s2
    """
    ee_inc = trapezoid(1 / rho_ray, 144 * 32.174 * prs_ray)
    return ee_inc


# change this to a function that just creates the book?
# this only goes past crossing the zero tde line
def throat_entry_zero_tde(
    psu: float, tsu: float, ken: float, ate: float, ipr_su: InFlow, prop_su: ResMix
) -> tuple[float, jp.JetBook]:
    """Throat Entry Differential Energy at Zero

    Create a throat entry book where the differential energy crosses zero.
    Throat entry book can
    Use IPR to find the expected well production rate at the specific psu.

    Args:
        psu (float): Suction Pressure, psig
        tsu (float): Suction Temp, deg F
        ken (float): Throat Entry Friction, unitless
        ate (float): Throat Entry Area, ft2
        ipr_su (InFlow): IPR of Reservoir
        prop_su (ResMix): Properties of Suction Fluid

    Returns:
        qoil_std (float): Oil Rate, STBOPD
        te_book (JetBook): Book of values for inside the throat entry
    """
    qoil_std = ipr_su.oil_flow(psu, method="pidx")  # oil standard flow, bopd

    prop_su = prop_su.condition(psu, tsu)
    qtot = sum(prop_su.insitu_volm_flow(qoil_std))
    vte = sp.velocity(qtot, ate)

    te_book = jp.JetBook(psu, vte, prop_su.rho_mix(), prop_su.cmix(), enterance_ke(ken, vte))

    pdec = 25  # pressure decrease
    pmin = 50

    while (te_book.tde_ray[-1] > 0) and (te_book.prs_ray[-1] > pmin):
        pte = te_book.prs_ray[-1] - pdec

        prop_su = prop_su.condition(pte, tsu)
        qtot = sum(prop_su.insitu_volm_flow(qoil_std))
        vte = sp.velocity(qtot, ate)

        te_book.append(pte, vte, prop_su.rho_mix(), prop_su.cmix(), enterance_ke(ken, vte))

        # re-evaluate criteria for this...
        # ensures crossing zero tde while below mach limit
        # if (te_book.mach_ray[-1] > 1) and (te_book.tde_ray[-2] > 100):
        # raise ValueError(f"Suction Pressure of {psu} psig is too low. Select higher Psu.")

    # pte, vte, rho_te, mach_te = te_book.dete_zero()
    return qoil_std, te_book


# this goes until the mach number equals one
def throat_entry_mach_one(
    psu: float, tsu: float, ken: float, ate: float, ipr_su: InFlow, prop_su: ResMix
) -> tuple[float, float, jp.JetBook]:
    """Throat Entry Differential Energy at Mach One

    Find the numerical value of the dEte (differential energy throat entry) equation
    when the gradient (slope?) is a zero value. The physical meaning of the when the slope
    is zero is when the transistion from subsonic (mach < 1) to sonic (mach > 1) flow. A
    detailed derivation that shows the difference relationship to dEte and Ma value can
    be produced by Kaelin Ellis upon request

    Args:
        psu (float): Suction Pressure, psig
        tsu (float): Suction Temp, deg F
        ken (float): Enterance Friction Factor, unitless
        ate (float): Throat Entry Area, ft2
        ipr_su (InFlow): IPR of Reservoir
        prop_su (ResMix): Properties of Suction Fluid

    Returns:
        tde_fin (float): Total Differential Energy at Mach 1, ft2/s2
        qoil_std (float): Oil Produced at psu with set IPR, bopd
        te_book (JetBook): Book of values for inside the throat entry
    """
    qoil_std = ipr_su.oil_flow(psu, method="pidx")  # oil standard flow, bopd

    prop_su = prop_su.condition(psu, tsu)
    qtot = sum(prop_su.insitu_volm_flow(qoil_std))
    vte = sp.velocity(qtot, ate)

    te_book = jp.JetBook(psu, vte, prop_su.rho_mix(), prop_su.cmix(), enterance_ke(ken, vte))

    pdec = 25  # pressure decrease
    pmin = 50  # minimum pressure

    # keep mach under one, and pte above pmin, so it doesn't go negative
    while (te_book.mach_ray[-1] <= 1) and (te_book.prs_ray[-1] > pmin):
        pte = te_book.prs_ray[-1] - pdec

        prop_su = prop_su.condition(pte, tsu)
        qtot = sum(prop_su.insitu_volm_flow(qoil_std))
        vte = sp.velocity(qtot, ate)

        te_book.append(pte, vte, prop_su.rho_mix(), prop_su.cmix(), enterance_ke(ken, vte))

    if te_book.mach_ray[-1] >= 1:  # return nearest value instead of interpolating
        tde_fin = te_book.tde_ray[-2]
    else:
        tde_fin = te_book.tde_ray[-1]

    return tde_fin, qoil_std, te_book  # type: ignore


def psu_minimize(
    tsu: float, ken: float, ate: float, ipr_su: InFlow, prop_su: ResMix
) -> tuple[float, float, jp.JetBook]:
    """Minimize psu

    Find the smallest psu possible where the throat is choked. (Ma = 1)
    This psu is the theoretically smallest psu possible for a set jetpump and ipr combo.
    Even with an infinite amount of power fluid, you could not get below this psu.

    Args:
        tsu (float): Suction Temp, deg F
        ken (float): Throat Entry Friction, unitless
        ate (float): Throat Entry Area, ft2
        ipr_su (InFlow): IPR of Reservoir
        prop_su (ResMix): Properties of Suction Fluid

    Returns:
        psu_min (float): Suction Pressure Minimized, psig
        qoil_std (float): Oil Rate, STBOPD
        pte (float): Throat Entry Pressure, psig
        rho_te (float): Throat Entry Density, lbm/ft3
        vte (float): Throat Entry Velocity, ft/s
    """
    psu_list = [ipr_su.pres - 300, ipr_su.pres - 400]
    # store values of tee near mach=1 pressure
    tee_list = [
        throat_entry_mach_one(psu_list[0], tsu, ken, ate, ipr_su, prop_su)[0],
        throat_entry_mach_one(psu_list[1], tsu, ken, ate, ipr_su, prop_su)[0],
    ]

    psu_diff = 5  # criteria for when you've converged to an answer
    n = 0  # loop counter
    while abs(psu_list[-2] - psu_list[-1]) > psu_diff:
        psu_nxt = psu_secant(psu_list[-2], psu_list[-1], tee_list[-2], tee_list[-1])
        tee_nxt, qoil_std, te_book = throat_entry_mach_one(psu_nxt, tsu, ken, ate, ipr_su, prop_su)
        psu_list.append(psu_nxt)
        tee_list.append(tee_nxt)
        n = n + 1
        if n == 10:
            raise ValueError("Suction Pressure for Minimization did not converge")
    # pte, vte, rho_te, mach_te = te_book.dete_zero()
    return psu_list[-1], qoil_std, te_book


def psu_secant(psu1: float, psu2: float, dete1: float, dete2: float) -> float:
    """Next Suction Pressure with Secant Method

    Uses the secant method to calculate the next psu to use to find a zero dEte at Ma = 1.

    Args:
        psu1 (float): Suction Pressure One, psig
        psu2 (float): Suction Pressure Two, psig
        dete1 (float): Differential Energy at psu1 and Ma = 1, ft2/s2
        dete2 (float): Differential Energy at psu1 and Ma = 1, ft2/s2

    Return:
        psu3 (float): Suction Pressure Three, psig
    """
    psu3 = psu2 - dete2 * (psu1 - psu2) / (dete1 - dete2)
    return psu3


def nozzle_velocity(pni: float, pte: float, knz: float, rho_nz: float) -> float:
    """Nozzle Velocity

    Solve Bernoulli's Equation to calculate the nozzle velocity in ft/s.

    Args:
        pni (float): Nozzle Inlet Pressure, psig
        pte (float): Throat Entry Pressure, psig
        knz (float): Friction of Nozzle, unitless
        rho_nz (float): Density of Nozzle Fluid, lbm/ft3

    Returns:
        vnz (float): Nozzle Velocity, ft/s
    """
    vnz = math.sqrt(2 * 32.174 * 144 * (pni - pte) / (rho_nz * (1 + knz)))
    return vnz


def nozzle_rate(vnz: float, anz: float) -> tuple[float, float]:
    """Nozzle Flow Rate

    Find Nozzle / Power Fluid Flowrate in ft3/s and BPD

    Args:
        vnz (float): Nozzle Velocity, ft/s
        anz (float): Area of Nozzle, ft2

    Returns:
        qnz_ft3s (float): Nozzle Flowrate ft3/s
        qnz_bpd (float): Nozzle Flowrate bpd
    """
    qnz_ft3s = anz * vnz
    qnz_bpd = sp.ft3s_to_bpd(qnz_ft3s)
    return qnz_ft3s, qnz_bpd


def throat_inlet_momentum(
    vnz: float, anz: float, rho_nz: float, vte: float, ate: float, rho_te: float
) -> tuple[float, float]:
    """Throat Inlet Momentum

    Calculate the inlet momentum of the throat in lbm/(s2*ft). The units
    are actually pressure, but referred to as momentum to help differentiate.

    Args:
        vnz (float): Velocity of Nozzle, ft/s
        anz (float): Area of Nozzle, ft2
        rho_nz (float): Density of Nozzle Fluid, lbm/ft3
        vte (float): Velocity of Throat Entry Mixture, ft/s
        ate (float): Area of Throat Entry, ft2
        rho_te (float): Density of Throat Entry Mixture, lbm/ft3

    Returns:
        mom_nz (float): Nozzle Momentum, lbm*ft/s2
        mom_te (float): Entry Momentum, lbm*ft/s2
    """
    mom_nz = sp.momentum(rho_nz, vnz, anz)  # momentum of the nozzle
    mom_te = sp.momentum(rho_te, vte, ate)  # momentum of the throat entry
    return mom_nz, mom_te


def throat_outlet_momentum(kth: float, vtm: float, ath: float, rho_tm: float) -> tuple[float, float]:
    """Throat Outlet Momentum

    Calculate the outlet momentum of the throat in lbm/(s2*ft). The units
    are actually pressure, but referred to as momentum to help differentiate.

    Args:
        kth (float): Throat Friction Factor, unitless
        vtm (float): Velocity of Throat Mixture, ft/s
        ath (float): Area of the Throat, ft2
        rho_tm (float): Density of Throat Mixture, lbm/ft3

    Returns:
        mom_tm (float): Throat Mixting Momentum, lbm*ft/s2
        mom_fr (float): Throat Friction Momemum, lbm*ft/s2
    """
    mom_tm = sp.momentum(rho_tm, vtm, ath)
    mom_fr = 1 / 2 * kth * mom_tm
    return mom_tm, mom_fr


def throat_discharge(
    pte: float,
    tte: float,
    kth: float,
    vnz: float,
    anz: float,
    rho_nz: float,
    vte: float,
    ate: float,
    rho_te: float,
    prop_tm: ResMix,
):
    """Throat Discharge Pressure

    Solves the throat mixture equation of the jet pump. Calculates throat differntial pressure.
    Use the throat entry pressure and differential pressure to calculate throat mix pressure.
    Account for the discharge pressure is greater than the inlet pressure. Loops through the
    calculated discharge pressure until a converged answer occurs.

    Args:
        pte (float): Pressure of Throat Entry, psig
        tte (float): Temperature of Throat Entry, deg F
        kth (float): Friction of Throat Mix, Unitless
        vnz (float): Velocity of Nozzle, ft/s
        anz (float): Area of Nozzle, ft2
        rho_nz (float): Density of Nozzle Fluid, lbm/ft3
        vte (float): Velocity of Throat Entry Mixture, ft/s
        ate (float): Area of Throat Entry, ft2
        rho_te (float): Density of Throat Entry Mixture, lbm/ft3
        prop_tm (ResMix): Properties of the Throat Mixture

    Returns:
        ptm (float): Throat Discharge Pressure, psig
    """
    mom_nz, mom_te = throat_inlet_momentum(vnz, anz, rho_nz, vte, ate, rho_te)

    mnz = sp.massflow(rho_nz, vnz, anz)  # mass flow of the nozzle
    mte = sp.massflow(rho_te, vte, ate)  # mass flow of the throat entry
    ath = anz + ate  # area of the throat
    mtm = mnz + mte  # mass flow of total mixture

    rho_tm = prop_tm.condition(pte, tte).rho_mix()  # density of total mixture
    vtm = sp.velocity(mtm / rho_tm, ath)

    mom_tm, mom_fr = throat_outlet_momentum(kth, vtm, ath, rho_tm)
    mom_tot = mom_fr + mom_tm - mom_nz - mom_te
    dp_tm = sp.mom_to_psi(mom_tot, ath)  # lbf/in2

    ptm = pte - dp_tm
    ptm_list = [pte, ptm]

    ptm_diff = 5
    n = 0
    while abs(ptm_list[-2] - ptm_list[-1]) > ptm_diff:
        ptm = ptm_list[-1]

        rho_tm = prop_tm.condition(ptm, tte).rho_mix()  # density of total mixture
        vtm = sp.velocity(mtm / rho_tm, ath)

        mom_tm, mom_fr = throat_outlet_momentum(kth, vtm, ath, rho_tm)
        mom_tot = mom_fr + mom_tm - mom_nz - mom_te
        dp_tm = sp.mom_to_psi(mom_tot, ath)  # lbf/in2
        ptm = pte - dp_tm
        ptm_list.append(ptm)
        n += 1
        if n == 10:
            raise ValueError("Throat Mixture did not converge")
    return ptm_list[-1]


def throat_wc(qoil_std: float, wc_su: float, qwat_nz: float) -> tuple[float, float]:
    """Throat Watercut and Formation Water Rate

    Calculate watercut and formation water rate into the jet pump throat.
    New watercut after the power fluid and reservoir fluid have mixed together.

    Args:
        qoil_std (float): Oil Rate, STD BOPD
        wc_su (float): Watercut at pump suction, decimal
        qwat_nz (float): Powerfluid Flowrate, BWPD

    Returns:
        wc_tm (float): Watercut at throat, decimal
        qwat_su (float): Formation Water at Suction, bwpd"""

    qwat_su = qoil_std * wc_su / (1 - wc_su)
    qwat_tot = qwat_nz + qwat_su
    wc_tm = qwat_tot / (qwat_tot + qoil_std)
    return wc_tm, qwat_su


def diffuser_ke(kdi: float, vtm: float, vdi: float) -> float:
    """Diffuser Kinetic Energy

    Calculate the kinetic energy in the diffuser.
    Substract the energy lost due to friction from inlet fluid.

    Args:
        kdi: Diffuser Friction Factor, unitless
        vtm: Velocity at Throat Mixture, ft/s
        vdi: Velocity at Diffuser Discharge, ft/s

    Returns:
        ke_di: Diffuser Kinetic Energy, ft2/s2
    """
    ke_di = (vdi**2 - (1 - kdi) * vtm**2) / 2
    return ke_di


def diffuser_discharge(
    ptm: float, ttm: float, kdi: float, ath: float, adi: float, qoil_std: float, prop_tm: ResMix
) -> tuple[float, float]:
    """Diffuser Discharge Pressure

    Directly calculate the diffuser discharge pressure. Only loops until the diffuser total
    energy is greater than 0. Then uses numpy interpolation for diffuser discharge pressure.

    Args:
        ptm (float): Throat Mixture Pressure, psig
        ttm (float): Throat Mixture Temp, deg F
        kdi (float): Diffuser Friction Factor, unitless
        ath (float): Throat Area, ft2
        adi (float): Diffuser / Tubing Area, ft2
        qoil_std (float): Oil Rate, STD BOPD
        prop_tm (ResMix): Properties of Throat Mixture

    Returns:
        vtm (float): Throat Mixture Velocity
        pdi (float): Diffuser Discharge Pressure, psig
    """
    prop_tm = prop_tm.condition(ptm, ttm)
    qtot = sum(prop_tm.insitu_volm_flow(qoil_std))
    vdi = sp.velocity(qtot, adi)
    vtm = sp.velocity(qtot, ath)

    di_book = jp.JetBook(ptm, vdi, prop_tm.rho_mix(), prop_tm.cmix(), diffuser_ke(kdi, vtm, vdi))

    pinc = 100  # pressure increase

    while di_book.tde_ray[-1] < 0:
        pdi = di_book.prs_ray[-1] + pinc

        prop_tm = prop_tm.condition(pdi, ttm)
        qtot = sum(prop_tm.insitu_volm_flow(qoil_std))
        vdi = sp.velocity(qtot, adi)

        di_book.append(pdi, vdi, prop_tm.rho_mix(), prop_tm.cmix(), diffuser_ke(kdi, vtm, vdi))

    pdi = di_book.dedi_zero()
    return vtm, pdi  # type: ignore


def jetpump_overall(
    psu: float,
    tsu: float,
    pni: float,
    rho_ni: float,
    ken: float,
    knz: float,
    kth: float,
    kdi: float,
    ath: float,
    anz: float,
    adi: float,
    ipr_su: InFlow,
    prop_su: ResMix,
) -> tuple[float, float, float, float, float, float, float, ResMix]:
    """Jet Pump Overall Equations

    Solve the jetpump equations, calculating out the expected discharge conditions.
    Function dete_zero() will raise a ValueError if the selected psu is too low.

    Args:
        psu (float): Suction Pressure, psig
        tsu (float): Suction Temp, deg F
        pni (float): Nozzle Inlet Pressure, psig
        rho_ni (float): Nozzle Inlet Density, lbm/ft3
        ken (float): Enterance Friction Factor, unitless
        knz (float): Nozzle Friction Factor, unitless
        kth (float): Throat Friction Factor, unitless
        kdi (float): Diffuser Friction Factor, unitless
        ath (float): Throat Area, ft2
        anz (float): Nozzle Area, ft2
        adi (float): Diffuser Area, ft2
        ipr_su (InFlow): IPR of Reservoir
        prop_su (ResMix): Properties of Suction Fluid

    Returns:
        pte (float): Throat Entry Pressure, psig
        ptm (float): Throat Mixture Pressure, psig
        pdi (float): Diffuser Discharge Pressure, psig
        qoil_std (float): Oil Rate, STBOPD
        fwat_bpd (float): Formation Water Rate, BWPD
        qnz_bpd (float): Power Fluid Rate, BWPD
        mach_te (float): Throat Entry Mach, unitless
        prop_tm (ResMix): Properties of Discharge Fluid
    """
    ate = ath - anz
    qoil_std, te_book = throat_entry_zero_tde(psu=psu, tsu=tsu, ken=ken, ate=ate, ipr_su=ipr_su, prop_su=prop_su)
    pte, vte, rho_te, mach_te = te_book.dete_zero()

    vnz = nozzle_velocity(pni, pte, knz, rho_ni)

    qnz_ft3s, qnz_bwpd = nozzle_rate(vnz, anz)
    wc_tm, fwat_bwpd = throat_wc(qoil_std, prop_su.wc, qnz_bwpd)

    prop_tm = ResMix(wc_tm, prop_su.fgor, prop_su.oil, prop_su.wat, prop_su.gas)
    ptm = throat_discharge(pte, tsu, kth, vnz, anz, rho_ni, vte, ate, rho_te, prop_tm)
    vtm, pdi = diffuser_discharge(ptm, tsu, kdi, ath, adi, qoil_std, prop_tm)
    return pte, ptm, pdi, qoil_std, fwat_bwpd, qnz_bwpd, mach_te, prop_tm



"""Simple Pump

Series of functions that don't use classes as inputs to simplify the required runs.
"""

from woffl.assembly import sysops as so
from woffl.flow import jetcheck as jc
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.blackoil import BlackOil
from woffl.pvt.formgas import FormGas
from woffl.pvt.formwat import FormWater
from woffl.pvt.resmix import ResMix


def jetpump_wrapper(
    is_sch: bool,
    pwh: float,
    rho_pf: float,
    ppf_surf: float,
    out_dia: float,
    thick: float,
    qwf: float,
    pwf: float,
    res_pres: float,
    nozzle_no: str,
    throat: str,
    form_wc: float,
    form_gor: float,
    form_temp: float,
    wellname: str,
):
    """Jet Pump Model Wrapper

    Wrapper function for jetpump_solver. Eliminates creating each class separately.

    Args:
        is_sch (boolean): Is the Well Schrader?
        pwh (float): Wellhead Pressure, psi
        rho_pf (float): Power fluid density, lbm/ft3
        ppf_surf (float): Power Fluid pressure at surface, psi
        out_dia (float): Tubing OD, inches
        thick (float): Tubing Wall thickness, inches
        qwf (float): Oil Rate at Flowing Bottom Hole Pressure, STBOPD
        pwf (float): Flowing Bottom Hole Pressure, psi
        res_pres (float): Reservoir Pressure, psig
        nozzle_no (str): Nozzle Size
        throat (str): Throat Ratio
        form_wc(float): Formation Water Cut, fraction
        form_gor (float): Formation Gas Oil Ratio, SCF/STB
        form_temp (float): Formation Temperature, deg F
        wellname (str): Name of Modeled Well

    Returns:
        psu_solv (float): Suction Pressure, psig
        sonic_status (boolean): Choked or not
        qoil_std (float): Oil Rate at Predicted Suction Pressure, STBOPD
        fwat_bwpd (float): Formation Water Rate, BWPD
        qnz_bwpd (float): Power Fluid Rate, BWPD
        mach_te(float): Mach Number Throat Entry, unitless
        total_wc (float): Total Water Cut (PF + Form), fraction
        total_water (float): Total Water Rate (PF + Form), BWPD
        wellname (str): Wellname
    """
    if is_sch:
        mpu_oil = BlackOil.schrader()  # class method
        mpu_wat = FormWater.schrader()  # class method
        mpu_gas = FormGas.schrader()  # class method

        wellprof = WellProfile.schrader()

    else:
        mpu_oil = BlackOil.kuparuk()  # class method
        mpu_wat = FormWater.kuparuk()  # class method
        mpu_gas = FormGas.kuparuk()  # class method

        wellprof = WellProfile.kuparuk()

    tube = Pipe(out_dia, thick)

    ipr_su = InFlow(qwf, pwf, res_pres)

    prop_su = ResMix(wc=form_wc, fgor=form_gor, oil=mpu_oil, wat=mpu_wat, gas=mpu_gas)

    jp = JetPump(nozzle_no, throat, ken=0.03, kth=0.3, kdi=0.4)

    psu_solv, sonic_status, qoil_std, fwat_bwpd, qnz_bwpd, mach_te = so.jetpump_solver(
        pwh, form_temp, rho_pf, ppf_surf, jp, tube, wellprof, ipr_su, prop_su
    )
    total_water = qnz_bwpd + fwat_bwpd
    total_wc = (qnz_bwpd + fwat_bwpd) / (qnz_bwpd + fwat_bwpd + qoil_std)

    print(f"\nResults for well: {wellname} ")
    print(f"JP Size: {nozzle_no} {throat}")
    print(f"Suction pressure: {psu_solv:.0f}")
    print(f"Sonic Status: {sonic_status}")
    print(f"Oil Rate: {qoil_std:.0f}")
    print(f"Water Rate: {fwat_bwpd:.0f}")
    print(f"PF Rate: {qnz_bwpd:.0f}")
    print(f"Mach TE: {mach_te:.2f}")
    print(f"Total WC: {total_wc:.2f}\n")

    return psu_solv, sonic_status, qoil_std, fwat_bwpd, qnz_bwpd, mach_te, total_wc, total_water, wellname



import matplotlib.pyplot as plt
import numpy as np

import woffl.flow.outflow as of
from woffl.geometry.pipe import Pipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.blackoil import BlackOil
from woffl.pvt.formgas import FormGas
from woffl.pvt.formwat import FormWater
from woffl.pvt.resmix import ResMix

# only works if the command python -m tests.outflow_test is used
# mirror the hysys stuff
md_list = np.linspace(0, 6000, 100)
vd_list = np.linspace(0, 4000, 100)

# well with 600 fgor, 90% wc, 100 bopd
mpu_oil = BlackOil.schrader_oil()
mpu_wat = FormWater.schrader_wat()
mpu_gas = FormGas.schrader_gas()
form_gor = 600  # scf/stb
form_wc = 0.9
qoil_std = 100  # stbopd

test_prop = ResMix(form_wc, form_gor, mpu_oil, mpu_wat, mpu_gas)
wellprof = WellProfile(md_list, vd_list, 6000)
tubing = Pipe(out_dia=4.5, thick=0.5)

ptop = 350  # psig
ttop = 100  # deg f

md_seg, prs_ray, slh_ray = of.top_down_press(ptop, ttop, qoil_std, test_prop, tubing, wellprof)

slh_ray = np.append(slh_ray, np.nan)  # add a nan to make same length for graphing
fig, ax1 = plt.subplots()
ax2 = ax1.twinx()
ax1.plot(md_seg, prs_ray, linestyle="--", color="b", label="Pressure")
ax1.set_ylabel("Pressure, PSIG")
ax1.set_xlabel("Measured Depth, Feet")
ax2.plot(md_seg, slh_ray, linestyle="-", color="r", label="Holdup")
ax2.set_ylabel("Slip Liquid Holdup")
lines1, labels1 = ax1.get_legend_handles_labels()
lines2, labels2 = ax2.get_legend_handles_labels()
ax2.legend(lines1 + lines2, labels1 + labels2, loc=0)
plt.show()

print(f"Bottom Pressure: {round(prs_ray[-1], 2)} psi")



import woffl.flow.singlephase as sp
import woffl.flow.twophase as tp

# only works if the command python -m tests.flow_test is used
# example problem in Two Phase Flow in Pipes by Beggs / Brill (1988) pg. 3-31
book_Ngv = 9.29
book_Nlv = 6.02
book_Nd = 41.34
book_l1, book_l2 = 1.53, 0.88
book_vpat = "slug"

calc_l1, calc_l2 = tp.ros_lp(book_Nd)
calc_vpat = tp.ros_flow_pattern(9.29, 6.02, 41.34)

print(f"Ros L1 & L2 - Book: {book_l1}, {book_l2}, Calc: {round(calc_l1, 2)}, {round(calc_l2, 2)}")
print(f"Ros Regime - Book: {book_vpat}, Calc: {calc_vpat}")

# example problem in Two Phase Flow in Pipes by Beggs / Brill (1988) pg. 3-62
book_nslh = 0.393
book_NFr = 5.67
book_hpat = "intermittent"
book_tparm = 1  # this is really n/a since it is intermittent flow
book_ilh = 0.512
book_dhyd = 0.249 * 12  # inches
book_rho_liq = 56.6
book_rho_gas = 2.84
book_uliq = 18
book_ugas = 0.018
book_NRe = 8450
book_vsg = 4.09  # ft/s
book_vsl = 2.65  # ft/s
book_vmix = book_vsg + book_vsl
book_ff = 0.032  # book doesn't say what value they use for absolute roughness
book_yf = 1.5
book_sf = 0.37
book_ftp = 0.046
pipe_len = 100  # feet
book_fric = 3.12 * pipe_len / 144  # convert from psf/ft to psi
book_stat = 30.36 * pipe_len / 144  # convert from psf/ft to psi

# function is the same for slip vs no slip, just depends on what value for slip vs no slip you use
book_rho_mix = tp.density_slip(book_rho_liq, book_rho_gas, book_nslh)
book_rho_slip = tp.density_slip(book_rho_liq, book_rho_gas, book_ilh)
book_umix = tp.density_slip(book_uliq, book_ugas, book_nslh)

calc_NFr = tp.froude(book_vmix, book_dhyd)
calc_hpat, calc_tparm = tp.beggs_flow_pattern(book_nslh, book_NFr)
calc_ilh = tp.beggs_holdup_inc(book_nslh, book_NFr, book_Nlv, 90, calc_hpat, calc_tparm)
calc_NRe = sp.reynolds(book_rho_mix, book_vmix, book_dhyd, book_umix)
calc_rr = sp.relative_roughness(book_dhyd, 0.004)  # book doesn't say what abs ruff they use...
calc_ff = sp.ffactor_darcy(calc_NRe, calc_rr)
calc_yf = tp.beggs_yf(book_nslh, book_ilh)
calc_sf = tp.beggs_sf(book_yf)
calc_ftp = tp.beggs_ff(book_ff, book_sf)
calc_fric = tp.beggs_press_friction(book_ftp, book_rho_mix, book_vmix, book_dhyd, pipe_len)
calc_stat = tp.beggs_press_static(book_rho_slip, pipe_len)

print(f"Froude Number - Book: {book_NFr}, Calc: {round(calc_NFr, 2)}")
print(f"Beggs Regime - Book: {book_hpat}, Calc: {calc_hpat}")
print(f"Beggs Incline Liquid Holdup - Book: {book_ilh}, Calc: {round(calc_ilh, 3)}")
print(f"Reynolds Number - Book: {book_NRe}, Calc: {round(calc_NRe, 1)}")
print(f"Friction Factor - Book: {book_ff}, Calc: {round(calc_ff, 3)}")
print(f"Beggs Y Factor - Book: {book_yf}, Calc: {round(calc_yf, 3)}")
print(f"Beggs S Factor - Book: {book_sf}, Calc: {round(calc_sf, 4)}")
print(f"Beggs Friction Factor - Book: {book_ftp}, Calc: {round(calc_ftp, 3)}")
print(f"Beggs Friction Drop - Book: {round(book_fric, 3)} psi, Calc: {round(calc_fric, 3)} psi")
print(f"Beggs Static Drop - Book: {round(book_stat, 3)} psi, Calc: {round(calc_stat, 3)} psi")



class DeadOil():

    def __init__(self, oil_api, bubblepoint):
        # define an oil stream that has no live gas in it
        # what it wad before it died

        self.oil_api = oil_api
        self.pbp = bubblepoint

    # does something when you print your class
    def __repr__(self):
        return f'Dead Oil: {self.oil_api} API and {self.pbp} PSIG BubblePoint'

# note, the problem with inheritance is that it inherits your method names
# so if BlackOil inherits properties from FormGas, when I call density, it
# will call the FormGas Density Properties instead of...what I want?
